import traceback
from io import BytesIO
from typing import (
    Any,
    Dict,
    List,
    Tuple,
)

from . import (
    BaseFileSystem,
    DirListResult,
    FileSystemOperationError,
    Path,
)


class S3FileSystem(BaseFileSystem):
    _s3_client: Any
    _protocol: str
    _separator: str

    def __init__(self, s3_client: Any, *, uri_protocol: str="s3", separator: str="/") -> None:
        self._s3_client = s3_client
        self._protocol = uri_protocol
        self._separator = separator

    # General operations
    def parse_path(self, path: str) -> Path:
        bucket_path_separator_position: int = path.find("/")
        drive: str = path[:bucket_path_separator_position]
        tail: List[str] = path[bucket_path_separator_position + 1:].split(self._separator)
        return Path(drive, *tail)

    def path_exists(self, path: Path) -> bool:
        prefix: str = self._separator.join(path.tail)
        response: Dict[str, List[Dict[str, str]]] = self._s3_client.list_objects_v2(
            Bucket=path.drive,
            Prefix=prefix,
            Delimiter=self._separator
        )

        contents: List[Dict[str, str]] = response.get('Contents', [])
        for item in contents:
            file_name: str = item['Key']
            if file_name == prefix:
                return True

        prefixes: List[Dict[str, str]] = response.get('CommonPrefixes', [])
        for item in prefixes:
            dir_name: str = item['Prefix']
            if dir_name == prefix + self._separator:
                return True

        return False

    def path_to_string(self, path: Path) -> str:
        return f"{path.drive}/{self._separator.join(path.tail)}"

    def path_to_uri(self, path: Path) -> str:
        return f"{self._protocol}://{self.path_to_string(path)}"

    # File operations
    def file_read(self, path: Path) -> bytes:
        prefix: str = self._separator.join(path.tail)
        try:
            data = BytesIO()
            self._s3_client.download_fileobj(path.drive, prefix, data)
            data.seek(0)
            return data.read()
        except Exception:
            raise FileSystemOperationError(traceback.format_exc())

    def file_write(self, path: Path, content: bytes) -> None:
        try:
            dirs: Tuple[str, ...] = path.parent.tail
            acc_prefix: str = ""
            for dir_ in dirs:
                if acc_prefix:
                    acc_prefix = self._separator.join((acc_prefix, dir_))
                else:
                    acc_prefix = dir_
                self._s3_client.upload_fileobj(BytesIO(b""), path.drive, acc_prefix + self._separator)

            path_str: str = self.path_to_string(path)
            tail: str = path_str[path_str.find("/") + 1:]
            self._s3_client.upload_fileobj(BytesIO(content), path.drive, tail)
        except Exception:
            raise FileSystemOperationError(traceback.format_exc())

    def file_remove(self, path: Path) -> None:
        prefix: str = self._separator.join(path.tail)
        try:
            response: Dict[str, List[Dict[str, str]]] = self._s3_client.list_objects_v2(
                Bucket=path.drive,
                Prefix=prefix,
            )
            contents: List[Dict[str, str]] = response.get('Contents', [])
            if contents:
                self._s3_client.delete_objects(
                    Bucket=path.drive,
                    Delete={
                        'Objects': [{'Key': f['Key']} for f in contents if f['Key'] == prefix]
                    }
                )
        except Exception:
            raise FileSystemOperationError(traceback.format_exc())

    # Directory operations
    def dir_list(self, path: Path) -> DirListResult:
        tail_str: str = self._separator.join(path.tail)
        delimiter: str = self._separator
        if tail_str:
            tail_str += self._separator
        try:
            response: Dict[str, List[Dict[str, str]]] = self._s3_client.list_objects_v2(
                Bucket=path.drive,
                Prefix=tail_str,
                Delimiter=delimiter
            )
        except Exception:
            raise FileSystemOperationError(traceback.format_exc())

        files: List[Path] = []
        dirs: List[Path] = []

        contents: List[Dict[str, str]] = response.get('Contents', [])
        for item in contents:
            file_name: str = item['Key']
            if file_name != tail_str:
                if file_name.endswith(self._separator):
                    dirs.append(Path(path.drive, *(file_name.split(self._separator)[:-1])))
                else:
                    files.append(Path(path.drive, *file_name.split(self._separator)))

        prefixes: List[Dict[str, str]] = response.get('CommonPrefixes', [])
        for item in prefixes:
            dir_name: str = item['Prefix']
            dirs.append(Path(path.drive, *(dir_name.split(self._separator)[:-1])))

        return DirListResult(files, dirs, [])

    def dir_remove(self, path: Path) -> None:
        try:
            response: Dict[str, List[Dict[str, str]]] = self._s3_client.list_objects_v2(
                Bucket=path.drive,
                Prefix=self._separator.join(path.tail) + self._separator
            )
            contents: List[Dict[str, str]] = response.get('Contents', [])
            if contents:
                self._s3_client.delete_objects(
                    Bucket=path.drive,
                    Delete={
                        'Objects': [{'Key': f['Key']} for f in contents]
                    }
                )
        except Exception:
            raise FileSystemOperationError(traceback.format_exc())

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}(s3_client={self._s3_client}, uri_protocol={self._protocol}, "
                f"separator={self._separator})")



import traceback

from collections import deque
from typing import (
    Dict,
    List,
)

from . import (
    BaseFileSystem,
    DirListResult,
    FileSystemOperationError,
    NotAFileError,
    Path,
)


class MemoryDirectory:
    _subdirs: Dict[str, 'MemoryDirectory']
    _files: Dict[str, 'MemoryFile']

    def __init__(self) -> None:
        self._subdirs = {}
        self._files = {}

    def list_dirs(self) -> List[str]:
        return list(self._subdirs.keys())

    def list_files(self) -> List[str]:
        return list(self._files.keys())

    def put_dir(self, name: str, dir_: 'MemoryDirectory') -> None:
        self._subdirs[name] = dir_

    def put_file(self, name: str, file_: 'MemoryFile') -> None:
        self._files[name] = file_

    def get_dir(self, name: str) -> 'MemoryDirectory':
        try:
            return self._subdirs[name]
        except KeyError:
            raise NotADirectoryError()

    def get_file(self, name: str) -> 'MemoryFile':
        try:
            return self._files[name]
        except KeyError:
            raise NotAFileError()

    def remove_dir(self, name: str) -> None:
        try:
            del self._subdirs[name]
        except KeyError:
            raise NotADirectoryError()

    def remove_file(self, name: str) -> None:
        try:
            del self._files[name]
        except KeyError:
            raise NotADirectoryError()


class MemoryFile:
    _contents: bytes

    def __init__(self, contents: bytes) -> None:
        self._contents = contents

    @property
    def contents(self) -> bytes:
        return self._contents


class MemoryFileSystem(BaseFileSystem):
    _root: MemoryDirectory
    _separator: str

    def __init__(self, root: MemoryDirectory, *, separator: str="/") -> None:
        self._root = root
        self._separator = separator

    def _get_dir(self, path: Path) -> MemoryDirectory:
        current_dir: MemoryDirectory = self._root
        for dir_name in path.tail:
            current_dir = current_dir.get_dir(dir_name)
        return current_dir

    # General operations
    def parse_path(self, path: str) -> Path:
        tail: List[str] = path.split(self._separator)
        return Path("", *tail)

    def path_exists(self, path: Path) -> bool:
        try:
            parent_dir: MemoryDirectory = self._get_dir(path.parent)
        except NotADirectoryError:
            return False
        else:
            return path.basename in (parent_dir.list_dirs() + parent_dir.list_files())

    def path_to_string(self, path: Path) -> str:
        return self._separator.join(path.tail)

    def path_to_uri(self, path: Path) -> str:
        return f"memory://{self.path_to_string(path)}"

    # File operations
    def file_read(self, path: Path) -> bytes:
        parent_dir: MemoryDirectory
        try:
            parent_dir = self._get_dir(path.parent)
            return parent_dir.get_file(path.basename).contents
        except (NotADirectoryError, NotAFileError):
            raise FileSystemOperationError(traceback.format_exc())

    def file_write(self, path: Path, content: bytes) -> None:
        parent_dir: MemoryDirectory = self._root
        dir_name: str
        new_dir: MemoryDirectory

        queue = deque(path.parent.tail)
        while queue:
            dir_name = queue.popleft()
            try:
                parent_dir = parent_dir.get_dir(dir_name)
            except NotADirectoryError:
                new_dir = MemoryDirectory()
                parent_dir.put_dir(dir_name, new_dir)
                parent_dir = new_dir
                break

        while queue:
            dir_name = queue.popleft()
            new_dir = MemoryDirectory()
            parent_dir.put_dir(dir_name, new_dir)
            parent_dir = new_dir

        parent_dir.put_file(path.basename, MemoryFile(content))

    def file_remove(self, path: Path) -> None:
        try:
            parent_dir: MemoryDirectory = self._get_dir(path.parent)
            parent_dir.remove_file(path.basename)
        except (NotADirectoryError, NotAFileError):
            raise FileSystemOperationError(traceback.format_exc())

    # Directory operations
    def dir_list(self, path: Path) -> DirListResult:
        files: List[Path]
        dirs: List[Path]
        current_dir: MemoryDirectory
        try:
            current_dir = self._get_dir(path)
        except NotADirectoryError:
            files = []
            dirs = []
        else:
            files = [path.child(file_) for file_ in current_dir.list_files()]
            dirs = [path.child(dir_) for dir_ in current_dir.list_dirs()]

        return DirListResult(files, dirs, [])

    def dir_remove(self, path: Path) -> None:
        try:
            current_dir: MemoryDirectory = self._get_dir(path.parent)
            current_dir.remove_dir(path.basename)
        except NotADirectoryError:
            raise FileSystemOperationError(traceback.format_exc())

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(root={self._root})"



from typing import (
    Any,
    List,
    Tuple,
)


class NotAFileError(Exception):
    pass


class FileSystemOperationError(Exception):
    pass


class Path:
    _drive: str
    _tail: Tuple[str, ...]

    def __init__(self, drive: str, *tail: str) -> None:
        self._drive = drive
        self._tail = tail

    @property
    def drive(self) -> str:
        return self._drive

    @property
    def tail(self) -> Tuple[str, ...]:
        return self._tail

    @property
    def basename(self) -> str:
        return self._tail[-1]

    @property
    def parent(self) -> 'Path':
        return Path(self._drive, *self._tail[:-1])

    def child(self, name: str) -> 'Path':
        return Path(self._drive, *self._tail, name)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(drive={self._drive.__repr__()}, tail={self._tail})"

    def __hash__(self) -> int:
        return hash(self.__repr__())

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Path):
            return self._drive == other.drive and self._tail == other.tail
        return False


class DirListResult:
    _files: List[Path]
    _dirs: List[Path]
    _others: List[Path]

    def __init__(self, files: List[Path], dirs: List[Path], others: List[Path]) -> None:
        self._files = files
        self._dirs = dirs
        self._others = others

    @property
    def files(self) -> List[Path]:
        return self._files

    @property
    def dirs(self) -> List[Path]:
        return self._dirs

    @property
    def others(self) -> List[Path]:
        return self._others

    def __eq__(self, other: Any) -> bool:
        if other is self:
            return True
        if not isinstance(other, DirListResult):
            return False
        if set(other.files) != set(self.files):
            return False
        if set(other.dirs) != set(self.dirs):
            return False
        if set(other.others) != set(self.others):
            return False
        return True

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(files={self._files}, dirs={self._dirs}, others={self._others})"


class BaseFileSystem:
    ERROR_MESSAGE = "BaseFileSystem: operation not implemented"

    # General operations
    def parse_path(self, path: str) -> Path:
        raise FileSystemOperationError(self.ERROR_MESSAGE)

    def path_exists(self, path: Path) -> bool:
        raise FileSystemOperationError(self.ERROR_MESSAGE)

    def path_to_string(self, path: Path) -> str:
        raise FileSystemOperationError(self.ERROR_MESSAGE)

    def path_to_uri(self, path: Path) -> str:
        raise FileSystemOperationError(self.ERROR_MESSAGE)

    # File operations
    def file_read(self, path: Path) -> bytes:
        raise FileSystemOperationError(self.ERROR_MESSAGE)

    def file_write(self, path: Path, content: bytes) -> None:
        raise FileSystemOperationError(self.ERROR_MESSAGE)

    def file_remove(self, path: Path) -> None:
        raise FileSystemOperationError(self.ERROR_MESSAGE)

    # Directory operations
    def dir_list(self, path: Path) -> DirListResult:
        raise FileSystemOperationError(self.ERROR_MESSAGE)

    def dir_remove(self, path: Path) -> None:
        raise FileSystemOperationError(self.ERROR_MESSAGE)

    # Other
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __eq__(self, other: Any) -> bool:
        return hash(self) == hash(other)

    def __hash__(self) -> int:
        return hash(id(self))



import traceback

from typing import (
    Any,
    Dict,
    Tuple,
    cast,
)

from ..fs import (
    BaseFileSystem,
    DirListResult,
    Path,
)


class CopyError(Exception):
    pass


class CopyFileSystemObject:
    _fs: BaseFileSystem
    _path: Path

    def __init__(self, fs: BaseFileSystem, path: Path) -> None:
        self._fs = fs
        self._path = path

    @property
    def fs(self) -> BaseFileSystem:
        return self._fs

    @property
    def path(self) -> Path:
        return self._path

    def copy(self, dst: 'CopyFileSystemObject', copier: 'GenericCopier') -> None:
        raise TypeError("Cannot copy from filesystem object that is not file or directory")

    def copy_from_file(self, src: 'CopyFile', copier: 'GenericCopier') -> None:
        raise TypeError("Cannot copy to filesystem object that is not file or directory")

    def copy_from_dir(self, src: 'CopyDirectory', copier: 'GenericCopier') -> None:
        raise TypeError("Cannot copy to filesystem object that is not file or directory")

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            other_casted: 'CopyFileSystemObject' = cast(CopyFileSystemObject, other)
            return self._fs == other_casted._fs and self._path == other_casted._path

        return False

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(fs={self._fs}, path={self._path})"


class CopyFile(CopyFileSystemObject):

    def copy(self, dst: 'CopyFileSystemObject', copier: 'GenericCopier') -> None:
        dst.copy_from_file(self, copier)

    def copy_from_file(self, src: 'CopyFile', copier: 'GenericCopier') -> None:
        copier.copy_file_to_file(src, self)

    def copy_from_dir(self, src: 'CopyDirectory', copier: 'GenericCopier') -> None:
        raise TypeError("Cannot copy Directory into a File.")


class CopyDirectory(CopyFileSystemObject):

    def file(self, suffix: str) -> 'CopyFile':
        return CopyFile(self._fs, self._path.child(suffix))

    def subdir(self, suffix: str) -> 'CopyDirectory':
        return CopyDirectory(self._fs, self._path.child(suffix))

    def copy(self, dst: 'CopyFileSystemObject', copier: 'GenericCopier') -> None:
        dst.copy_from_dir(self, copier)

    def copy_from_file(self, src: 'CopyFile', copier: 'GenericCopier') -> None:
        copier.copy_file_to_file(src, self.file(self.path.basename))

    def copy_from_dir(self, src: 'CopyDirectory', copier: 'GenericCopier') -> None:
        copier.copy_dir_to_dir(src, self)


class GenericCopier:

    def copy_dir_to_dir(self, src: CopyDirectory, dst: CopyDirectory) -> None:
        contents: DirListResult = src.fs.dir_list(src.path)

        for file_ in contents.files:
            src_child_file: CopyFile = src.file(file_.basename)
            dst_child_file: CopyFile = dst.file(file_.basename)
            self.copy_file_to_file(src_child_file, dst_child_file)

        for dir_ in contents.dirs:
            src_child_dir: CopyDirectory = src.subdir(dir_.basename)
            dst_child_dir: CopyDirectory = dst.subdir(dir_.basename)
            self.copy_dir_to_dir(src_child_dir, dst_child_dir)

    def copy_file_to_file(self, src: CopyFile, dst: CopyFile) -> None:
        dst.fs.file_write(dst.path, src.fs.file_read(src.path))


class CopyHandler:
    _copiers: Dict[Tuple[BaseFileSystem, BaseFileSystem], GenericCopier]
    _default: GenericCopier

    def __init__(self, default_copier: GenericCopier) -> None:
        self._copiers = {}
        self._default = default_copier

    def set_copy_policy(self, src_fs: BaseFileSystem, dst_fs: BaseFileSystem, copier: GenericCopier) -> None:
        self._copiers[(src_fs, dst_fs)] = copier

    def copy(self, src: CopyFileSystemObject, dst: CopyFileSystemObject) -> None:
        copier: GenericCopier = self._copiers.get((src.fs, dst.fs), self._default)

        try:
            src.copy(dst, copier)
        except Exception:
            raise CopyError(traceback.format_exc())



import os
import shutil
import traceback

from typing import List

from . import (
    BaseFileSystem,
    DirListResult,
    FileSystemOperationError,
    Path,
)


class LocalFileSystem(BaseFileSystem):

    # General operations
    def parse_path(self, path: str) -> Path:
        abs_path: str = os.path.abspath(os.path.normpath(path))
        drive: str
        tail_str: str
        drive, tail_str = os.path.splitdrive(abs_path)
        tail: List[str] = tail_str.split(os.sep)
        return Path(drive, *tail)

    def path_exists(self, path: Path) -> bool:
        return os.path.exists(self.path_to_string(path))

    def path_to_string(self, path: Path) -> str:
        return os.path.join(path.drive, os.sep.join(path.tail))

    def path_to_uri(self, path: Path) -> str:
        return "file:///{}/{}".format(path.drive, os.path.join(*path.tail))

    # File operations
    def file_read(self, path: Path) -> bytes:
        try:
            with open(self.path_to_string(path), "rb") as f:
                return f.read()
        except Exception:
            raise FileSystemOperationError(traceback.format_exc())

    def file_write(self, path: Path, content: bytes) -> None:
        try:
            parent_path: str = self.path_to_string(path.parent)
            if not os.path.exists(parent_path):
                os.makedirs(parent_path)
            with open(self.path_to_string(path), "wb") as f:
                f.write(content)
        except Exception:
            raise FileSystemOperationError(traceback.format_exc())

    def file_remove(self, path: Path) -> None:
        try:
            os.remove(self.path_to_string(path))
        except Exception:
            raise FileSystemOperationError(traceback.format_exc())

    # Directory operations
    def dir_list(self, path: Path) -> DirListResult:
        path_str: str = self.path_to_string(path)
        items: List[str]
        try:
            items = os.listdir(path_str)
        except FileNotFoundError:
            items = []
        files: List[Path] = []
        dirs: List[Path] = []
        others: List[Path] = []
        for item in items:
            full_path: str = os.path.join(path_str, item)
            item_path: Path = path.child(item)
            if os.path.isfile(full_path):
                files.append(item_path)
            elif os.path.isdir(full_path):
                dirs.append(item_path)
            else:
                others.append(item_path)

        return DirListResult(files, dirs, others)

    def dir_remove(self, path: Path) -> None:
        try:
            shutil.rmtree(self.path_to_string(path))
        except Exception:
            raise FileSystemOperationError(traceback.format_exc())



from typing import (
    Any,
    Dict,
    Optional,
)

from .copy import (
    CopyHandler,
    GenericCopier,
)
from .copy.local import (
    LocalToLocalCopier,
    LocalToS3Copier,
)
from .copy.s3 import (
    S3ToS3Copier,
    S3ToLocalCopier,
)
from .filesystem import (
    Directory,
    File,
)
from .fs.local import LocalFileSystem
from .fs.s3 import S3FileSystem
from .fs.memory import (
    MemoryFileSystem,
    MemoryDirectory,
)


_config: Optional[Dict[str, Any]] = None
_copy_handler: Optional[CopyHandler] = None
_local_fs: Optional[LocalFileSystem] = None
_s3_fs: Optional[S3FileSystem] = None
_memory_fs: Optional[MemoryFileSystem] = None


class Version:
    MAJOR: str = "1"
    MINOR: str = "2"
    PATCH: str = "1"

    VERSION: str = ".".join((MAJOR, MINOR))
    RELEASE: str = ".".join((MAJOR, MINOR, PATCH))


def get_copy_handler() -> CopyHandler:
    """ Returns the global instance of CopyHandler. """
    global _copy_handler

    if _copy_handler is None:
        _copy_handler = CopyHandler(GenericCopier())

    return _copy_handler


def _get_config() -> Dict[str, Any]:
    global _config

    if _config is None:
        _config = {
            "s3_client": None,
            "s3_protocol": "s3",
            "s3_separator": "/",
            "memory_separator": "/",
        }

    return _config


def configure(**kwargs: Any) -> None:
    """ Set the configuration for the library.
    Accepted keyword values:

    *  `s3_client`: boto3 S3 client instance for S3FileSystem.
    *  `s3_protocol`: protocol to use for S3FileSystem uri. Defaults to "s3".
    *  `s3_separator`: directory separator to use for S3FileSystem. Defaults to "/".
    *  `memory_separator`: directory separator to use for MemoryFileSystem. Defaults to "/".
    """
    _get_config().update(kwargs)


def _init_local_fs() -> LocalFileSystem:
    global _local_fs

    _local_fs = LocalFileSystem()

    get_copy_handler().set_copy_policy(_local_fs, _local_fs, LocalToLocalCopier())

    return _local_fs


def get_local_fs() -> LocalFileSystem:
    """ Returns the global instance of LocalFileSystem. """
    global _local_fs

    return _local_fs or _init_local_fs()


def localdir(path: str) -> Directory:
    """ Return a Directory instance with the given `path` for the LocalFileSystem. """
    return Directory(get_local_fs(), path, get_copy_handler())


def localfile(path: str) -> File:
    """ Return a File instance with the given `path` for the LocalFileSystem. """
    return File(get_local_fs(), path, get_copy_handler())


def _init_s3_fs() -> S3FileSystem:
    global _s3_fs

    config: Dict[str, Any] = _get_config()

    if config["s3_client"] is None:
        raise ValueError("Cannot initialize S3FileSystem if s3_client is not set.\n"
                         "You can set it using `norfs.configure(s3_client=boto3.client(\"s3\"))`")

    s3_client: Any = config["s3_client"]

    _s3_fs = S3FileSystem(s3_client, uri_protocol=config["s3_protocol"], separator=config["s3_separator"])

    copy_handler: CopyHandler = get_copy_handler()
    copy_handler.set_copy_policy(get_local_fs(), _s3_fs, LocalToS3Copier(s3_client))
    copy_handler.set_copy_policy(_s3_fs, get_local_fs(), S3ToLocalCopier(s3_client))
    copy_handler.set_copy_policy(_s3_fs, _s3_fs, S3ToS3Copier(s3_client))

    return _s3_fs


def get_s3_fs() -> S3FileSystem:
    """ Returns the global instance of S3FileSystem. """
    global _s3_fs

    return _s3_fs or _init_s3_fs()


def _get_s3_path(path_or_bucket: str, prefix: Optional[str]=None) -> str:
    if prefix is None:
        return path_or_bucket
    return "/".join((path_or_bucket, prefix))


def s3dir(path_or_bucket: str, prefix: Optional[str]=None) -> Directory:
    """ Return a Directory instance with the given bucket and path for the S3FileSystem.

    If only one parameter is given the bucket name will be everything up to the first "/".
    """
    return Directory(get_s3_fs(), _get_s3_path(path_or_bucket, prefix), get_copy_handler())


def s3file(path_or_bucket: str, prefix: Optional[str]=None) -> File:
    """ Return a File instance with the given bucket and path for the S3FileSystem.

    If only one parameter is given the bucket name will be everything up to the first "/".
    """
    return File(get_s3_fs(), _get_s3_path(path_or_bucket, prefix), get_copy_handler())


def _init_memory_fs() -> MemoryFileSystem:
    global _memory_fs

    config: Dict[str, Any] = _get_config()

    _memory_fs = MemoryFileSystem(MemoryDirectory(), separator=config["memory_separator"])

    return _memory_fs


def get_memory_fs() -> MemoryFileSystem:
    """ Returns the global instance of MemoryFileSystem. """
    global _memory_fs

    return _memory_fs or _init_memory_fs()


def memorydir(path: str) -> Directory:
    """ Return a Directory instance with the given `path` for the MemoryFileSystem. """
    return Directory(get_memory_fs(), path, get_copy_handler())


def memoryfile(path: str) -> File:
    """ Return a File instance with the given `path` for the MemoryFileSystem. """
    return File(get_memory_fs(), path, get_copy_handler())



from setuptools import setup, find_packages
from norfs import Version

setup(name="norfs",
      version=Version.RELEASE,
      author="Alessio Linares",
      author_email="alessio@alessio.cc",
      description=("Normalized filesystem. This library offers a common interface to interact with multiple "
                   "filesystems, local or remote."),
      keywords=["filesystem", "boto", "s3", "file", "directory", "norfs"],
      url="https://github.com/Galbar/norfs",
      download_url="https://github.com/Galbar/norfs/archive/{}.tar.gz".format(Version.RELEASE),
      license='MIT',
      classifiers=[
          # How mature is this project? Common values are
          #   3 - Alpha
          #   4 - Beta
          #   5 - Production/Stable
          'Development Status :: 5 - Production/Stable',

          # Indicate who your project is intended for
          'Intended Audience :: Developers',
          'Topic :: Software Development :: Libraries',

          # Pick your license as you wish (should match "license" above)
          'License :: OSI Approved :: MIT License',

          # Specify the Python versions you support here. In particular, ensure
          # that you indicate whether you support Python 2, Python 3 or both.
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
      ],
      python_requires='>=3.6',
      packages=find_packages(exclude=["tests*"]))



import os

from typing import Any

from . import (
    GenericCopier,
    CopyFile,
)
from ..fs import Path


class S3ToS3Copier(GenericCopier):

    def __init__(self, s3_client: Any) -> None:
        self._s3_client = s3_client

    def copy_file_to_file(self, src: CopyFile, dst: CopyFile) -> None:
        src_path_str: str = src.fs.path_to_string(src.path)
        src_tail: str = src_path_str[src_path_str.find("/") + 1:]
        copy_source = {
            'Bucket': src.path.drive,
            'Key': src_tail
        }
        dst_path_str: str = dst.fs.path_to_string(dst.path)
        dst_tail: str = dst_path_str[dst_path_str.find("/") + 1:]
        self._s3_client.copy(copy_source, dst.path.drive, dst_tail)


class S3ToLocalCopier(GenericCopier):

    def __init__(self, s3_client: Any) -> None:
        self._s3_client = s3_client

    def copy_file_to_file(self, src: CopyFile, dst: CopyFile) -> None:
        parent_dir: Path = dst.path.parent
        if not dst.fs.path_exists(parent_dir):
            os.makedirs(dst.fs.path_to_string(parent_dir))
        src_path_str: str = src.fs.path_to_string(src.path)
        src_tail: str = src_path_str[src_path_str.find("/") + 1:]
        self._s3_client.download_file(src.path.drive, src_tail, dst.fs.path_to_string(dst.path))



from typing import (
    Any,
    List,
    Optional,
    cast,
)

from .fs import (
    BaseFileSystem,
    DirListResult,
    FileSystemOperationError,
    NotAFileError,
    Path,
)
from .copy import CopyHandler
from .copy import (
    CopyDirectory,
    CopyFile,
    CopyFileSystemObject,
)


class BaseFileSystemObject:
    _copy_handler: CopyHandler
    _fs: BaseFileSystem
    _path: Path

    def __init__(self, filesystem: BaseFileSystem, path_str: Optional[str], copy_handler: CopyHandler, *,
                 _path: Optional[Path]=None) -> None:
        """ Constructor for BaseFileSystemObjects.
        One of `path_str` and `_path` **MUST** be present.
        """
        self._copy_handler = copy_handler
        self._fs = filesystem
        self._path = _path or self._fs.parse_path(path_str or "")

    @property
    def path(self) -> str:
        """ The full, absolute, path of self in the file system. """
        return self._fs.path_to_string(self._path)

    @property
    def uri(self) -> str:
        """ The URI that points to self in the file system. """
        return self._fs.path_to_uri(self._path)

    @property
    def name(self) -> str:
        """ The name of self. """
        return self._path.basename

    def is_file(self) -> bool:
        """ Returns wether self is a File. """
        return False

    def is_dir(self) -> bool:
        """ Returns wether self is a Directory. """
        return False

    def as_file(self) -> 'File':
        """ Returns itself as a File instance or raises a NotAFileError. """
        raise NotAFileError()

    def as_dir(self) -> 'Directory':
        """ Returns itself as a Directory instance or raises a NotADirectoryError. """
        raise NotADirectoryError()

    def exists(self) -> bool:
        """ Returns whether self exists in the file system. """
        return self._fs.path_exists(self._path)

    def remove(self) -> None:
        """ Tries to remove self from the file system.
        On failure it raises a FileSystemOperationError
        """
        raise FileSystemOperationError(f"Cannot remove {str(self)}")

    def parent(self) -> 'Directory':
        """ Return parent Directory of self. """
        return Directory(self._fs, None, self._copy_handler, _path=self._path.parent)

    def copy(self, destination: 'BaseFileSystemObject') -> None:
        """ Copy this to `destination`. """
        self._copy_handler.copy(self._copy_object(), destination._copy_object())

    def _copy_object(self) -> CopyFileSystemObject:
        return CopyFileSystemObject(self._fs, self._path)

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}(fs={self._fs}, path={self.path}, copy_handler={self._copy_handler})")

    def __hash__(self) -> int:
        return hash(self.__repr__())

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            other_casted: 'BaseFileSystemObject' = cast(BaseFileSystemObject, other)
            return (self._path == other_casted._path and
                    self._fs == other_casted._fs and
                    self._copy_handler == other_casted._copy_handler)
        return False


class Directory(BaseFileSystemObject):

    def is_dir(self) -> bool:
        """ Returns wether self is a Directory. """
        return True

    def as_dir(self) -> 'Directory':
        """ Returns itself as a Directory instance or raises a NotADirectoryError. """
        return self

    def list(self) -> List[BaseFileSystemObject]:
        """ Returns the contents of the Directory in the file system as a list of BaseFileSystemObjects.

        If the Directory does not exist the list will be empty.
        """
        contents: DirListResult = self._fs.dir_list(self._path)
        result: List[BaseFileSystemObject] = []
        for dir_path in contents.dirs:
            result.append(Directory(self._fs, None, self._copy_handler, _path=dir_path))
        for file_path in contents.files:
            result.append(File(self._fs, None, self._copy_handler, _path=file_path))
        for other_path in contents.others:
            result.append(BaseFileSystemObject(self._fs, None, self._copy_handler, _path=other_path))

        return result

    def remove(self) -> None:
        """ Tries to remove self from the file system.

        On failure it raises a FileSystemOperationError
        """
        self._fs.dir_remove(self._path)

    def subdir(self, path: str) -> 'Directory':
        """ Returns a Directory with its path as being the given path relative to the current Directory. """
        return Directory(self._fs, None, self._copy_handler, _path=self._path.child(path))

    def file(self, path: str) -> 'File':
        """ Returns a File with its path as being the given `path` relative to the current Directory. """
        return File(self._fs, None, self._copy_handler, _path=self._path.child(path))

    def _copy_object(self) -> CopyFileSystemObject:
        return CopyDirectory(self._fs, self._path)


class File(BaseFileSystemObject):

    def is_file(self) -> bool:
        """ Returns wether self is a File. """
        return True

    def as_file(self) -> 'File':
        """ Returns itself as a File instance or raises a NotAFileError. """
        return self

    def remove(self) -> None:
        """ Tries to remove self from the file system.

        On failure it raises a FileSystemOperationError
        """
        self._fs.file_remove(self._path)

    def read(self) -> bytes:
        """ Returns the contents of the File.

        If it fails to read the file a FileSystemOperationError will be raised.
        """
        return self._fs.file_read(self._path)

    def write(self, content: bytes) -> None:
        """ Sets the contents of the File. If the parent directory does not exist it is created.

        If it fails to read the file a FileSystemOperationError will be raised.
        """
        self._fs.file_write(self._path, content)

    def _copy_object(self) -> CopyFileSystemObject:
        return CopyFile(self._fs, self._path)
